import { BigNumber, Contract, Event, EventFilter, providers } from "ethers";
import winston from "winston";
import { AnyObject, DefaultLogLevels, EventSearchConfig, MakeOptional } from "../utils";
import { Deposit, DepositWithBlock, DepositWithBlockStringified, Fill, FillWithBlock, FillWithBlockStringified, FundsDepositedEvent, FundsDepositedEventStringified, RefundRequestWithBlock, RefundRequestWithBlockStringified, RelayerRefundExecutionWithBlock, RelayerRefundExecutionWithBlockStringified, RootBundleRelayWithBlock, SpeedUp, SpeedUpStringified, TokensBridged, TokensBridgedStringified } from "../interfaces";
import { BaseAbstractClient } from "./BaseAbstractClient";
import { HubPoolClient } from "./HubPoolClient";
type Block = providers.Block;
type _SpokePoolUpdate = {
    success: boolean;
    currentTime: number;
    firstDepositId: number;
    latestBlockNumber: number;
    latestDepositId: number;
    events: Event[][];
    blocks: {
        [blockNumber: number]: Block;
    };
    searchEndBlock: number;
};
export type SpokePoolUpdate = {
    success: false;
} | _SpokePoolUpdate;
export declare class SpokePoolClient extends BaseAbstractClient {
    readonly logger: winston.Logger;
    readonly spokePool: Contract;
    readonly hubPoolClient: HubPoolClient | null;
    readonly chainId: number;
    deploymentBlock: number;
    readonly eventSearchConfig: MakeOptional<EventSearchConfig, "toBlock">;
    protected currentTime: number;
    protected depositHashes: {
        [depositHash: string]: DepositWithBlock;
    };
    protected depositHashesToFills: {
        [depositHash: string]: FillWithBlock[];
    };
    protected speedUps: {
        [depositorAddress: string]: {
            [depositId: number]: SpeedUp[];
        };
    };
    protected depositRoutes: {
        [originToken: string]: {
            [DestinationChainId: number]: boolean;
        };
    };
    protected tokensBridged: TokensBridged[];
    protected rootBundleRelays: RootBundleRelayWithBlock[];
    protected relayerRefundExecutions: RelayerRefundExecutionWithBlock[];
    protected earlyDeposits: FundsDepositedEvent[];
    protected queryableEventNames: string[];
    earliestDepositIdQueried: number;
    latestDepositIdQueried: number;
    firstDepositIdForSpokePool: number;
    lastDepositIdForSpokePool: number;
    firstBlockToSearch: number;
    latestBlockSearched: number;
    latestBlockNumber: number;
    fills: {
        [OriginChainId: number]: FillWithBlock[];
    };
    refundRequests: RefundRequestWithBlock[];
    constructor(logger: winston.Logger, spokePool: Contract, hubPoolClient: HubPoolClient | null, chainId: number, deploymentBlock: number, eventSearchConfig?: MakeOptional<EventSearchConfig, "toBlock">);
    _queryableEventNames(): {
        [eventName: string]: EventFilter;
    };
    getDepositsForDestinationChain(destinationChainId: number): DepositWithBlock[];
    getDeposits(): DepositWithBlock[];
    getTokensBridged(): TokensBridged[];
    getDepositRoutes(): {
        [originToken: string]: {
            [DestinationChainId: number]: boolean;
        };
    };
    isDepositRouteEnabled(originToken: string, destinationChainId: number): boolean;
    getAllOriginTokens(): string[];
    getFills(): FillWithBlock[];
    getFillsForOriginChain(originChainId: number): FillWithBlock[];
    getFillsForRelayer(relayer: string): FillWithBlock[];
    getFillsWithBlockInRange(startingBlock: number, endingBlock: number): FillWithBlock[];
    getRefundRequests(fromBlock?: number, toBlock?: number): RefundRequestWithBlock[];
    getRootBundleRelays(): RootBundleRelayWithBlock[];
    getLatestRootBundleId(): number;
    getRelayerRefundExecutions(): RelayerRefundExecutionWithBlock[];
    getExecutedRefunds(relayerRefundRoot: string): {
        [tokenAddress: string]: {
            [relayer: string]: BigNumber;
        };
    };
    appendMaxSpeedUpSignatureToDeposit(deposit: DepositWithBlock): DepositWithBlock;
    getDepositForFill(fill: Fill, fillFieldsToIgnore?: string[]): DepositWithBlock | undefined;
    updateDepositRealizedLpFeePct(event: Deposit, realizedLpFeePct: BigNumber): void;
    getValidUnfilledAmountForDeposit(deposit: Deposit): {
        unfilledAmount: BigNumber;
        fillCount: number;
        invalidFills: Fill[];
    };
    getDepositHash(event: Deposit | Fill | DepositWithBlock | FillWithBlock): string;
    _getBlockRangeForDepositId(targetDepositId: number, initLow: number, initHigh: number, maxSearches: number): Promise<{
        low: number;
        high: number;
    }>;
    _getDepositIdAtBlock(blockTag: number): Promise<number>;
    queryHistoricalMatchingFills(fill: Fill, deposit: Deposit, toBlock: number): Promise<FillWithBlock[]>;
    queryFillsInBlockRange(matchingFill: Fill, searchConfig: EventSearchConfig): Promise<FillWithBlock[]>;
    protected _update(eventsToQuery: string[]): Promise<SpokePoolUpdate>;
    _isEarlyDeposit(depositEvent: FundsDepositedEvent, currentTime: number): boolean;
    update(eventsToQuery?: string[]): Promise<void>;
    static getExecutedRefundLeafL2Token(chainId: number, eventL2Token: string): string;
    protected computeRealizedLpFeePct(depositEvent: FundsDepositedEvent): Promise<{
        realizedLpFeePct: BigNumber | undefined;
        quoteBlock: number;
    }> | {
        realizedLpFeePct: BigNumber;
        quoteBlock: number;
    };
    protected getDestinationTokenForDeposit(deposit: DepositWithBlock): string;
    protected log(level: DefaultLogLevels, message: string, data?: AnyObject): void;
    getCurrentTime(): number;
    findDeposit(depositId: number, destinationChainId: number, depositor: string): Promise<DepositWithBlock>;
    updateFromJSON(spokePoolClientState: Partial<ReturnType<SpokePoolClient["toJSON"]>>): void;
    toJSON(): {
        chainId: number;
        deploymentBlock: number;
        eventSearchConfig: MakeOptional<EventSearchConfig, "toBlock">;
        currentTime: number;
        depositHashes: Record<string, DepositWithBlockStringified>;
        depositHashesToFills: Record<string, FillWithBlockStringified[]>;
        speedUps: {
            [depositorAddress: string]: {
                [depositId: number]: SpeedUpStringified[];
            };
        };
        depositRoutes: {
            [originToken: string]: {
                [DestinationChainId: number]: boolean;
            };
        };
        tokensBridged: TokensBridgedStringified[];
        rootBundleRelays: RootBundleRelayWithBlock[];
        relayerRefundExecutions: RelayerRefundExecutionWithBlockStringified[];
        earlyDeposits: FundsDepositedEventStringified[];
        queryableEventNames: string[];
        earliestDepositIdQueried: number;
        earliestDepositIdForSpokePool: number;
        latestDepositIdQueried: number;
        latestDepositIdForSpokePool: number;
        firstBlockToSearch: number;
        latestBlockSearched: number;
        isUpdated: boolean;
        fills: Record<string, FillWithBlockStringified[]>;
        refundRequests: RefundRequestWithBlockStringified[];
    };
}
export {};
