import { __assign, __awaiter, __generator, __spreadArray } from "tslib";
import assert from "assert";
import { BigNumber, ethers } from "ethers";
import UBAFeeConfig from "../../UBAFeeCalculator/UBAFeeConfig";
import { mapAsync } from "../../utils/ArrayUtils";
import { isDefined } from "../../utils/TypeGuards";
import { validateFillForDeposit } from "../../utils/FlowUtils";
import { isUbaInflow, isUbaOutflow, outflowIsFill, } from "../../interfaces";
import { getBlockForChain, getBlockRangeForChain, getImpliedBundleBlockRanges } from "../../utils/BundleUtils";
import { stringifyJSONWithNumericString } from "../../utils/JSONUtils";
import { toBN } from "../../utils";
/**
 * Omit the default key from a dictionary
 * @param obj The dictionary to omit the default key from
 * @returns The dictionary without the default key
 * @note This is used to omit the default key from the UBA config
 */
function omitDefaultKeys(obj) {
    return Object.keys(obj).reduce(function (acc, key) {
        var _a;
        if (key !== "default") {
            return __assign(__assign({}, acc), (_a = {}, _a[key] = obj[key], _a));
        }
        return acc;
    }, {});
}
/**
 * Parses a UBAParsedConfigType into a UBAFeeConfig
 * @param chainId The chain ID to parse the UBA config for
 * @param tokenSymbol The token symbol to parse the UBA config for
 * @param ubaConfig The UBAParsedConfigType to parse
 * @returns The parsed UBAFeeConfig
 */
export function parseUBAFeeConfig(chainId, tokenSymbol, ubaConfig) {
    var _a;
    if (!ubaConfig) {
        return undefined;
    }
    var omegaDefault = ubaConfig.omega["default"];
    var omegaOverride = omitDefaultKeys(ubaConfig.omega);
    var gammaDefault = ubaConfig.gamma["default"];
    var gammaOverride = omitDefaultKeys(ubaConfig.gamma);
    var threshold = ubaConfig.rebalance[String(chainId)];
    var chainTokenCombination = "".concat(chainId, "-").concat(tokenSymbol);
    return new UBAFeeConfig({
        default: ubaConfig.alpha["default"],
        override: omitDefaultKeys(ubaConfig.alpha),
    }, {
        default: omegaDefault,
        override: omegaOverride,
    }, {
        default: {
            lowerBound: {
                target: toBN(0),
                threshold: toBN(0),
            },
            upperBound: {
                target: toBN(0),
                threshold: toBN(0),
            },
        },
        override: (_a = {},
            _a[chainTokenCombination] = {
                lowerBound: {
                    target: threshold === null || threshold === void 0 ? void 0 : threshold.target_lower,
                    threshold: threshold === null || threshold === void 0 ? void 0 : threshold.threshold_lower,
                },
                upperBound: {
                    target: threshold === null || threshold === void 0 ? void 0 : threshold.target_upper,
                    threshold: threshold === null || threshold === void 0 ? void 0 : threshold.threshold_upper,
                },
            },
            _a),
    }, {
        default: gammaDefault,
        override: gammaOverride,
    }, ubaConfig.incentivePoolAdjustment, ubaConfig.ubaRewardMultiplier);
}
/**
 * Returns the UBA config for a given chainId and tokenSymbol at a given block height.
 * @param hubPoolClient The hub pool client to use for fetching the UBA config. This client's config store client must be updated.
 * @param chainId The chainId to fetch the UBA config for.
 * @param tokenSymbol The token symbol to fetch the UBA config for.
 * @param blockNumber The block height to fetch the UBA config for.
 * @returns The UBA config for the given chainId and tokenSymbol at the given block height.
 * @throws If the config store client has not been updated at least once.
 * @throws If the L1 token address cannot be found for the given token symbol.
 * @throws If the UBA config for the given block height cannot be found.
 */
export function getUBAFeeConfig(hubPoolClient, chainId, tokenSymbol, blockNumber) {
    var configClient = hubPoolClient.configStoreClient;
    // If the config client has not been updated at least
    // once, throw
    if (!configClient.isUpdated) {
        throw new Error("Config client not updated");
    }
    var l1TokenInfo = hubPoolClient.getL1Tokens().find(function (token) { return token.symbol === tokenSymbol; });
    if (!l1TokenInfo) {
        throw new Error("L1 token can't be found, have you updated hub pool client?");
    }
    var rawUbaConfig = configClient.getUBAConfig(l1TokenInfo === null || l1TokenInfo === void 0 ? void 0 : l1TokenInfo.address, blockNumber);
    var ubaConfig = parseUBAFeeConfig(chainId, tokenSymbol, rawUbaConfig);
    if (ubaConfig === undefined) {
        throw new Error("UBA config for blockTag ".concat(blockNumber, " not found"));
    }
    return ubaConfig;
}
/**
 * Returns most recent `maxBundleStates` bundle ranges for a given chain, in chronological ascending order.
 * Will only returns bundle ranges that are subject to UBA rules, which is based on the bundle's start block.
 * Additionally, includes potential next bundle block ranges which extend until the latest spoke pool client search
 * windows, so that the caller can call this function to cover all UBA events.
 * @param chainId
 * @param maxBundleStates If this is larger than available validated bundles in the HubPoolClient, will throw an error.
 * @param hubPoolClient The hub pool client to use for fetching the bundle ranges.
 * @param spokePoolClients The spoke pool clients to use for fetching the bundle ranges.
 * @returns The most recent `maxBundleStates` bundle ranges for a given chain, in chronological ascending order.
 */
export function getMostRecentBundleBlockRanges(chainId, maxBundleStates, hubPoolClient, spokePoolClients) {
    var toBlock = hubPoolClient.latestBlockNumber;
    if (!isDefined(toBlock)) {
        throw new Error("HubPoolClient has undefined latestBlockNumber");
    }
    // Reconstruct bundle ranges based on published end blocks.
    var ubaActivationStartBlocks = getUbaActivationBundleStartBlocks(hubPoolClient);
    var ubaActivationHubStartBlock = ubaActivationStartBlocks[0];
    // Bundle states are examined in chronological descending order.
    var bundleData = [];
    for (var i = 0; i < maxBundleStates; i++) {
        // Get the most recent bundle end range for this chain published in a bundle before `toBlock`.
        var latestExecutedRootBundle = hubPoolClient.getNthFullyExecutedRootBundle(-1, toBlock);
        if (!latestExecutedRootBundle) {
            break;
        }
        // If chain is inactive for this bundle, exit early since no subsequent bundles will be enabled.
        var chainIndex = hubPoolClient.configStoreClient.getChainIdIndicesForBlock().indexOf(chainId);
        if (chainIndex >= latestExecutedRootBundle.bundleEvaluationBlockNumbers.length) {
            break;
        }
        var rootBundleBlockRanges = getImpliedBundleBlockRanges(hubPoolClient, hubPoolClient.configStoreClient, latestExecutedRootBundle);
        // If UBA is not enabled for this bundle, exit early since no subsequent bundles will be enabled
        // for the UBA, as the UBA was a non-reversible change and we're only looking older in the past from here.
        var hubPoolStartBlock = getBlockRangeForChain(rootBundleBlockRanges, hubPoolClient.chainId, hubPoolClient.configStoreClient.getChainIdIndicesForBlock())[0];
        if (hubPoolStartBlock < ubaActivationHubStartBlock) {
            break;
        }
        // Push the block range for this chain to the start of the list
        var blockRangeForChain = getBlockRangeForChain(rootBundleBlockRanges, chainId, hubPoolClient.configStoreClient.getChainIdIndicesForBlock());
        bundleData.unshift({
            start: blockRangeForChain[0],
            end: blockRangeForChain[1],
        });
        // Decrement toBlock to the start of the most recently grabbed bundle.
        toBlock = latestExecutedRootBundle.blockNumber;
    }
    // If we haven't saved any bundles yet, then there probably hasn't been a bundle validated after the UBA activation
    // block containing this chain. This is not necessarily an error, so inject a block range from the activation block
    // of this chain to the latest spoke block searched so we can query all UBA eligible events.
    if (bundleData.length === 0) {
        var ubaActivationBundleStartBlocks = getUbaActivationBundleStartBlocks(hubPoolClient);
        var ubaActivationBundleStartBlockForChain = getBlockForChain(ubaActivationBundleStartBlocks, chainId, hubPoolClient.configStoreClient.getChainIdIndicesForBlock());
        bundleData.unshift({
            // Tell caller to load data for events beginning at the start of the UBA version added to the ConfigStore
            start: ubaActivationBundleStartBlockForChain,
            // We'll extend this end block if the spoke pool client for this chain is defined.
            end: ubaActivationBundleStartBlockForChain,
        });
    }
    if (isDefined(spokePoolClients[chainId])) {
        // Make the last bundle to cover until the last spoke client searched block, unless a spoke pool
        // client was provided for the chain. In this case we assume that chain is disabled.
        bundleData[bundleData.length - 1].end = spokePoolClients[chainId].latestBlockSearched;
    }
    return bundleData;
}
/**
 * Return the latest validated running balance for the given token.
 * @param eventBlock Block number on SpokePool chain.
 * @param eventChain SpokePool chain ID.
 * @param l1Token Address of token on the HubPool chain.
 * @param hubPoolLatestBlock Most recent block to query from. Queries progress backwards in reverse chronology.
 * @returns Token running balance as at eventBlock.
 **/
export function getOpeningRunningBalanceForEvent(hubPoolClient, eventBlock, eventChain, l1Token, hubPoolLatestBlock) {
    var chainIdIndices = hubPoolClient.configStoreClient.getChainIdIndicesForBlock();
    // First find the latest executed bundle as of `hubPoolLatestBlock`.
    var latestExecutedBundle = hubPoolClient.getNthFullyExecutedRootBundle(-1, hubPoolLatestBlock);
    // If there is no latest executed bundle, then return 0. This means that there is no
    // bundle before `hubPoolLatestBlock` containing the event block.
    if (!isDefined(latestExecutedBundle)) {
        return {
            runningBalance: ethers.constants.Zero,
            incentiveBalance: ethers.constants.Zero,
        };
    }
    // If chain was not active in bundle, return 0.
    var chainIndex = chainIdIndices.indexOf(eventChain);
    if (chainIndex >= latestExecutedBundle.bundleEvaluationBlockNumbers.length) {
        return {
            runningBalance: ethers.constants.Zero,
            incentiveBalance: ethers.constants.Zero,
        };
    }
    // Construct the bundle's block range
    var blockRanges = getImpliedBundleBlockRanges(hubPoolClient, hubPoolClient.configStoreClient, latestExecutedBundle);
    // Now compare the eventBlock against the eventBlockRange.
    var eventBlockRange = getBlockRangeForChain(blockRanges, eventChain, chainIdIndices);
    // If event block is after the bundle end block, use the running balances for this bundle. We need to enforce
    // that the bundle end block is less than the event block to ensure that the running balance from this bundle
    // precedes the event block.
    if (eventBlock > eventBlockRange[1]) {
        // This can't be empty since we've already validated that this bundle is fully executed.
        var executedLeavesForBundle = hubPoolClient.getExecutedLeavesForRootBundle(latestExecutedBundle, hubPoolLatestBlock);
        if (executedLeavesForBundle.length === 0) {
            throw new Error("No executed leaves found for bundle");
        }
        var executedLeaf = executedLeavesForBundle.find(function (executedLeaf) { return executedLeaf.chainId === eventChain; });
        if (!executedLeaf) {
            // If no executed leaf in this bundle for the chain, then need to look for an older executed bundle.
            return getOpeningRunningBalanceForEvent(hubPoolClient, eventBlock, eventChain, l1Token, latestExecutedBundle.blockNumber);
        }
        var l1TokenIndex = executedLeaf.l1Tokens.indexOf(l1Token);
        if (l1TokenIndex === -1) {
            // If l1 token not included in this bundle, then need to look for an older executed bundle.
            return getOpeningRunningBalanceForEvent(hubPoolClient, eventBlock, eventChain, l1Token, latestExecutedBundle.blockNumber);
        }
        // Finally, we need to do a final check if the latest executed root bundle was the final pre UBA one. If so, then
        // its running balances need to be negated, because in the pre UBA world we counted "positive balances" held
        // by spoke pools as negative running balances.
        var runningBalance = executedLeaf.runningBalances[l1TokenIndex];
        var incentiveBalance = executedLeaf.incentiveBalances[l1TokenIndex];
        var ubaActivationStartBlocks = getUbaActivationBundleStartBlocks(hubPoolClient);
        var ubaActivationStartBlockForChain = getBlockForChain(ubaActivationStartBlocks, eventChain, chainIdIndices);
        if (blockRanges[0][0] < ubaActivationStartBlockForChain) {
            return {
                runningBalance: runningBalance.mul(-1),
                // Incentive balance starts at 0.
                incentiveBalance: ethers.constants.Zero,
            };
        }
        else {
            return {
                runningBalance: runningBalance,
                incentiveBalance: incentiveBalance,
            };
        }
    }
    // Event is either in the bundle or before it, look for an older executed bundle.
    return getOpeningRunningBalanceForEvent(hubPoolClient, eventBlock, eventChain, l1Token, latestExecutedBundle.blockNumber);
}
export function getMatchedDeposit(spokePoolClients, fill, fillFieldsToIgnore) {
    if (fillFieldsToIgnore === void 0) { fillFieldsToIgnore = []; }
    return __awaiter(this, void 0, void 0, function () {
        var originSpokePoolClient, deposit;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    originSpokePoolClient = spokePoolClients[fill.originChainId];
                    // We need to update client so we know the first and last deposit ID's queried for this spoke pool client, as well
                    // as the global first and last deposit ID's for this spoke pool.
                    if (!originSpokePoolClient.isUpdated) {
                        throw new Error("SpokePoolClient must be updated before querying historical deposits");
                    }
                    // If someone fills with a clearly bogus deposit ID then we can quickly mark it as invalid
                    if (fill.depositId < originSpokePoolClient.firstDepositIdForSpokePool ||
                        fill.depositId > originSpokePoolClient.lastDepositIdForSpokePool) {
                        return [2 /*return*/, undefined];
                    }
                    if (fill.depositId >= originSpokePoolClient.earliestDepositIdQueried &&
                        fill.depositId <= originSpokePoolClient.latestDepositIdQueried) {
                        return [2 /*return*/, originSpokePoolClient.getDepositForFill(fill, fillFieldsToIgnore)];
                    }
                    return [4 /*yield*/, originSpokePoolClient.findDeposit(fill.depositId, fill.destinationChainId, fill.depositor)];
                case 1:
                    deposit = _a.sent();
                    return [2 /*return*/, validateFillForDeposit(fill, deposit, fillFieldsToIgnore) ? deposit : undefined];
            }
        });
    });
}
export function isUBAActivatedAtBlock(hubPoolClient, block, chain) {
    var _a;
    try {
        var ubaActivationBlocks = getUbaActivationBundleStartBlocks(hubPoolClient);
        var chainIdIndices_1 = hubPoolClient.configStoreClient.getChainIdIndicesForBlock();
        // Find the first activation block where the index matches the chain
        var activationBlock = (_a = ubaActivationBlocks.find(function (_, idx) { return chainIdIndices_1[idx] === chain; })) !== null && _a !== void 0 ? _a : Number.MAX_SAFE_INTEGER;
        return block >= activationBlock;
    }
    catch (err) {
        // UBA not activated yet or hub pool client not updated
        return false;
    }
}
/**
 * Returns bundle range start blocks for first bundle that UBA was activated
 * @param chainIds Chains to return start blocks for.
 * */
export function getUbaActivationBundleStartBlocks(hubPoolClient) {
    var latestHubPoolBlock = hubPoolClient.latestBlockNumber;
    if (!isDefined(latestHubPoolBlock)) {
        throw new Error("HubPoolClient has undefined latestBlockNumber");
    }
    var chainIdIndices = hubPoolClient.configStoreClient.getChainIdIndicesForBlock();
    var ubaActivationBlock = hubPoolClient.configStoreClient.getUBAActivationBlock();
    if (isDefined(ubaActivationBlock)) {
        var nextValidatedBundle_1 = hubPoolClient.getProposedRootBundles().find(function (bundle) {
            return bundle.blockNumber >= ubaActivationBlock
                ? hubPoolClient.isRootBundleValid(bundle, latestHubPoolBlock)
                : false;
        });
        if (isDefined(nextValidatedBundle_1)) {
            var bundleBlockRanges_1 = getImpliedBundleBlockRanges(hubPoolClient, hubPoolClient.configStoreClient, nextValidatedBundle_1);
            var bundleStartBlocks = chainIdIndices.map(function (chainId, i) {
                // If chain was not active in `nextValidatedBundle`, then return 0 as the start block. This chain's next
                // bundle start block will be 0.
                if (i >= nextValidatedBundle_1.bundleEvaluationBlockNumbers.length) {
                    return 0;
                }
                var startBlock = getBlockRangeForChain(bundleBlockRanges_1, chainId, chainIdIndices)[0];
                return startBlock;
            });
            return bundleStartBlocks;
        }
        else {
            // No validated bundles after UBA activation block, UBA should be activated on next bundle start blocks.
            var nextBundleStartBlocks = chainIdIndices.map(function (chainId) {
                return hubPoolClient.getNextBundleStartBlockNumber(chainIdIndices, latestHubPoolBlock, chainId);
            });
            return nextBundleStartBlocks;
        }
    }
    else {
        throw new Error("UBA was not activated yet as of ".concat(latestHubPoolBlock));
    }
}
/**
 * Return chain where flow occurred.
 * @param flow
 */
export function getFlowChain(flow) {
    // First let's figure out the balancing fee for the chain where the flow takes place.
    // For a deposit, the flow happens on the origin chain.
    // For a refund or fill, the flow happens on the repayment or destination chain.
    var flowChain;
    if (isUbaInflow(flow)) {
        flowChain = flow.originChainId;
    }
    else {
        if (outflowIsFill(flow)) {
            flowChain = flow.destinationChainId;
        }
        else {
            flowChain = flow.repaymentChainId;
        }
    }
    return flowChain;
}
/**
 * Retrieves the flows for a given chainId. This is designed only to work with UBA start and end blocks as
 * the realizedLpFeePct for deposits is not compared against fills. Retrieving any "Pre UBA" deposits with
 * defined realizedLpFeePct will cause this function to throw.
 * * @param tokenSymbol Symbol of token to retrieve flows for.
 * @param chainId The chainId to retrieve flows for
 * @param spokePoolClients A mapping of chainIds to spoke pool clients
 * @param hubPoolClient A hub pool client instance to query the hub pool
 * @param fromBlock The block number to start retrieving flows from
 * @param toBlock The block number to stop retrieving flows from
 * @returns The flows for the given chainId
 */
export function getUBAFlows(tokenSymbol, chainId, spokePoolClients, hubPoolClient, fromBlock, toBlock) {
    return __awaiter(this, void 0, void 0, function () {
        var spokePoolClient, deposits, preUBADeposit, fills, refundRequests, flows;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    spokePoolClient = spokePoolClients[chainId];
                    fromBlock = fromBlock !== null && fromBlock !== void 0 ? fromBlock : spokePoolClient.eventSearchConfig.fromBlock;
                    toBlock = toBlock !== null && toBlock !== void 0 ? toBlock : spokePoolClient.eventSearchConfig.toBlock;
                    deposits = spokePoolClient.getDeposits().filter(function (deposit) {
                        var _a;
                        if (deposit.blockNumber < fromBlock || deposit.blockNumber > toBlock) {
                            return false;
                        }
                        var _tokenSymbol = (_a = hubPoolClient.getL1TokenInfoForL2Token(deposit.originToken, deposit.originChainId)) === null || _a === void 0 ? void 0 : _a.symbol;
                        return _tokenSymbol === tokenSymbol;
                    });
                    preUBADeposit = deposits.find(function (deposit) { return isDefined(deposit.realizedLpFeePct); });
                    if (isDefined(preUBADeposit)) {
                        throw new Error("Found a UBA deposit with a defined realizedLpFeePct at block ".concat(preUBADeposit.blockNumber, " on chain ").concat(preUBADeposit.originChainId));
                    }
                    return [4 /*yield*/, getValidFillCandidates(chainId, spokePoolClients, {
                            fromBlock: fromBlock,
                            toBlock: toBlock,
                            repaymentChainId: chainId,
                            isSlowRelay: false,
                            isCompleteFill: true,
                        }, ["realizedLpFeePct"])];
                case 1:
                    fills = (_a.sent()).filter(function (fill) {
                        var _a;
                        var _tokenSymbol = (_a = hubPoolClient.getL1TokenInfoForL2Token(fill.destinationToken, fill.destinationChainId)) === null || _a === void 0 ? void 0 : _a.symbol;
                        // We only want to include full fills as flows. Partial fills need to request refunds and those refunds
                        // will be included as flows.
                        return fill.fillAmount.eq(fill.totalFilledAmount) && _tokenSymbol === tokenSymbol;
                    });
                    return [4 /*yield*/, getValidRefundCandidates(chainId, hubPoolClient, spokePoolClients, {
                            fromBlock: fromBlock,
                            toBlock: toBlock,
                        }, ["realizedLpFeePct"])];
                case 2:
                    refundRequests = (_a.sent()).filter(function (refundRequest) {
                        var _a;
                        var _tokenSymbol = (_a = hubPoolClient.getL1TokenInfoForL2Token(refundRequest.refundToken, refundRequest.repaymentChainId)) === null || _a === void 0 ? void 0 : _a.symbol;
                        return _tokenSymbol === tokenSymbol;
                    });
                    flows = sortFlowsAscending(deposits.concat(fills).concat(refundRequests));
                    return [2 /*return*/, flows];
            }
        });
    });
}
export function getBundleKeyForBlockRanges(blockRanges) {
    return JSON.stringify(blockRanges.map(function (blockRange) { return blockRange[0]; }));
}
/**
 * The return value should be a number whose sign indicates the relative order of the two elements:
 * negative if a is less than b, positive if a is greater than b, and zero if they are equal.
 * @param fx
 * @param fy
 * @returns
 */
export function flowComparisonFunction(a, b) {
    if (a.blockTimestamp !== b.blockTimestamp) {
        return a.blockTimestamp - b.blockTimestamp;
    }
    // If fx and fy have same blockTimestamp then it gets a bit arbitrary. Whatever
    // we implement here we should match with the UMIP.
    // In the case of inflow vs outflow, return inflow first:
    if (isUbaInflow(a) && isUbaOutflow(b)) {
        return -1;
    }
    else if (isUbaInflow(b) && isUbaOutflow(a)) {
        return 1;
    }
    // We can compare on transaction index and log index when comparing same way
    // flows on same chain (i.e. deposits with deposits, fills with fills,
    // refunds with refunds).
    if (a.blockNumber !== b.blockNumber) {
        return a.blockNumber - b.blockNumber;
    }
    if (a.transactionIndex !== b.transactionIndex) {
        return a.transactionIndex - b.transactionIndex;
    }
    return a.logIndex - b.logIndex;
}
export function sortFlowsAscendingInPlace(flows) {
    return flows.sort(function (fx, fy) { return flowComparisonFunction(fx, fy); });
}
export function sortFlowsAscending(flows) {
    return sortFlowsAscendingInPlace(__spreadArray([], flows, true));
}
/**
 * Validate a refund request.
 * @param chainId The chainId of the spoke pool
 * @param chainIdIndices The chainIds of the spoke pools that align with the spoke pool clients
 * @param spokePoolClients A mapping of chainIds to spoke pool clients
 * @param hubPoolClient The hub pool client
 * @param refundRequest The refund request to validate
 * @returns Whether or not the refund request is valid
 */
export function refundRequestIsValid(spokePoolClients, hubPoolClient, refundRequest, ignoredDepositValidationParams) {
    if (ignoredDepositValidationParams === void 0) { ignoredDepositValidationParams = []; }
    return __awaiter(this, void 0, void 0, function () {
        var relayer, amount, refundToken, depositId, originChainId, destinationChainId, repaymentChainId, realizedLpFeePct, fillBlock, previousIdenticalRequests, destSpoke, deploymentBlock, latestBlockNumber, fill, deposit, l1TokenForFill, expectedRefundToken;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    relayer = refundRequest.relayer, amount = refundRequest.amount, refundToken = refundRequest.refundToken, depositId = refundRequest.depositId, originChainId = refundRequest.originChainId, destinationChainId = refundRequest.destinationChainId, repaymentChainId = refundRequest.repaymentChainId, realizedLpFeePct = refundRequest.realizedLpFeePct, fillBlock = refundRequest.fillBlock, previousIdenticalRequests = refundRequest.previousIdenticalRequests;
                    if (destinationChainId === repaymentChainId) {
                        return [2 /*return*/, { valid: false, reason: "Invalid destinationChainId" }];
                    }
                    destSpoke = spokePoolClients[destinationChainId];
                    if (fillBlock.lt(destSpoke.deploymentBlock) || fillBlock.gt(destSpoke.latestBlockNumber)) {
                        deploymentBlock = destSpoke.deploymentBlock, latestBlockNumber = destSpoke.latestBlockNumber;
                        return [2 /*return*/, {
                                valid: false,
                                reason: "FillBlock (".concat(fillBlock, " out of SpokePool range [").concat(deploymentBlock, ", ").concat(latestBlockNumber, "]"),
                            }];
                    }
                    // @dev: In almost all cases we should only count refunds where this value is 0. However, sometimes its possible
                    // that an initial refund request is thrown out due to some odd timing bug so this might be overly restrictive.
                    if (previousIdenticalRequests.gt(0)) {
                        return [2 /*return*/, { valid: false, reason: "Previous identical request exists" }];
                    }
                    fill = destSpoke.getFillsForRelayer(relayer).find(function (fill) {
                        // prettier-ignore
                        return (fill.depositId === depositId
                            && fill.originChainId === originChainId
                            && fill.destinationChainId === destinationChainId
                            // Must have requested refund on chain that refund was sent on.
                            && fill.repaymentChainId === repaymentChainId
                            // Must be a full fill to qualify for a refund.
                            && fill.amount.eq(amount)
                            && fill.fillAmount.eq(amount)
                            && fill.realizedLpFeePct.eq(realizedLpFeePct)
                            && fill.blockNumber === fillBlock.toNumber());
                    });
                    if (!isDefined(fill)) {
                        if (fillBlock.lt(destSpoke.eventSearchConfig.fromBlock)) {
                            // TODO: We need to do a look back for a fill if we can't find it. We can't assume it doesn't exist, similar to
                            // why we try to do a longer lookback below for a deposit. The problem with looking up the fill is that there is no
                            // deterministic way to eliminate the possibility that a fill exists.
                            // However, its OK to assume this refund is invalid for now since we assume that refunds are sent very close
                            // to the time of the fill.
                            // Can try to use `getModifiedFlow` in combination with some new findFill method in the SpokePoolClient.
                            throw new Error("Unimplemented: refund request fillBlock ".concat(fillBlock, " is older than spoke pool client from block, set a wider lookback"));
                        }
                        else {
                            return [2 /*return*/, { valid: false, reason: "Unable to find matching fill" }];
                        }
                    }
                    return [4 /*yield*/, getMatchedDeposit(spokePoolClients, fill, ignoredDepositValidationParams)];
                case 1:
                    deposit = _a.sent();
                    if (!isDefined(deposit)) {
                        return [2 /*return*/, { valid: false, reason: "Unable to find matching deposit" }];
                    }
                    // Verify that the refundToken maps to a known HubPool token and is the correct
                    // token for the chain where the refund was sent from.
                    // Note: the refundToken must be valid at the time the deposit was sent.
                    try {
                        l1TokenForFill = hubPoolClient.getL1TokenCounterpartAtBlock(fill.destinationChainId, fill.destinationToken, deposit.quoteBlockNumber);
                        expectedRefundToken = hubPoolClient.getDestinationTokenForL1Token(l1TokenForFill, repaymentChainId);
                        if (expectedRefundToken !== refundToken) {
                            return [2 /*return*/, { valid: false, reason: "Refund token does not map to expected refund token ".concat(refundToken) }];
                        }
                    }
                    catch (_b) {
                        return [2 /*return*/, { valid: false, reason: "Refund token unknown at HubPool block ".concat(deposit.quoteBlockNumber) }];
                    }
                    return [2 /*return*/, { valid: true, matchingFill: fill, matchingDeposit: deposit }];
            }
        });
    });
}
/**
 * @notice Get the matching flow in a stream of already validated flows. Useful for seeing if an outflow's
 * matched inflow `targetFlow` is in the `allValidatedFlows` list.
 * @dev Assumes `allValidatedFlows` are all validated UBA flows so there should be no duplicate
 * origin chain and deposit id combinations.
 * @param allFlows
 * @param targetFlow
 * @returns
 */
export function getMatchingFlow(allValidatedFlows, targetFlow) {
    return allValidatedFlows === null || allValidatedFlows === void 0 ? void 0 : allValidatedFlows.find(function (_a) {
        var flow = _a.flow;
        return flow.depositId === targetFlow.depositId && flow.originChainId === targetFlow.originChainId;
    });
}
/**
 * @description Search for fills recorded by a specific SpokePool. These fills are matched against deposits
 * on all fields except for `realizedLpFeePct` which isn't filled yet.
 * @param chainId Chain ID of the relevant SpokePoolClient instance.
 * @param spokePoolClients Set of SpokePoolClient instances, mapped by chain ID.
 * @param filter  Optional filtering criteria.
 * @param ignoredDepositValidationParams Event fields to ignore when matching a Fill to a Deposit.
 * @returns Array of FillWithBlock events matching the chain ID and optional filtering criteria.
 */
export function getValidFillCandidates(chainId, spokePoolClients, filter, ignoredDepositValidationParams) {
    if (filter === void 0) { filter = {}; }
    if (ignoredDepositValidationParams === void 0) { ignoredDepositValidationParams = []; }
    return __awaiter(this, void 0, void 0, function () {
        var spokePoolClient, repaymentChainId, relayer, isSlowRelay, isCompleteFill, fromBlock, toBlock, fills;
        var _this = this;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    spokePoolClient = spokePoolClients[chainId];
                    assert(isDefined(spokePoolClient));
                    repaymentChainId = filter.repaymentChainId, relayer = filter.relayer, isSlowRelay = filter.isSlowRelay, isCompleteFill = filter.isCompleteFill, fromBlock = filter.fromBlock, toBlock = filter.toBlock;
                    return [4 /*yield*/, mapAsync(spokePoolClient.getFills(), function (fill) { return __awaiter(_this, void 0, void 0, function () {
                            var deposit;
                            return __generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0:
                                        if (isDefined(fromBlock) && fromBlock > fill.blockNumber) {
                                            return [2 /*return*/, undefined];
                                        }
                                        if (isDefined(toBlock) && toBlock < fill.blockNumber) {
                                            return [2 /*return*/, undefined];
                                        }
                                        // @dev tsdx and old Typescript seem to prevent dynamic iteration over the filter, so evaluate the keys manually.
                                        if ((isDefined(repaymentChainId) && fill.repaymentChainId !== repaymentChainId) ||
                                            (isDefined(relayer) && fill.relayer !== relayer) ||
                                            (isDefined(isSlowRelay) && fill.updatableRelayData.isSlowRelay !== isSlowRelay)) {
                                            return [2 /*return*/, undefined];
                                        }
                                        if (isDefined(isCompleteFill) && isCompleteFill !== fill.fillAmount.eq(fill.totalFilledAmount)) {
                                            return [2 /*return*/, undefined];
                                        }
                                        return [4 /*yield*/, getMatchedDeposit(spokePoolClients, fill, ignoredDepositValidationParams)];
                                    case 1:
                                        deposit = _a.sent();
                                        if (deposit !== undefined) {
                                            return [2 /*return*/, __assign(__assign({}, fill), { matchedDeposit: deposit })];
                                        }
                                        else
                                            return [2 /*return*/, undefined];
                                        return [2 /*return*/];
                                }
                            });
                        }); })];
                case 1:
                    fills = (_a.sent()).filter(isDefined);
                    return [2 /*return*/, fills];
            }
        });
    });
}
/**
 * Search for refund requests recorded by a specific SpokePool.
 * @param chainId Chain ID of the relevant SpokePoolClient instance.
 * @param chainIdIndices Complete set of ordered chain IDs.
 * @param hubPoolClient HubPoolClient instance.
 * @param spokePoolClients Set of SpokePoolClient instances, mapped by chain ID.
 * @param filter  Optional filtering criteria.
 * @returns Array of RefundRequestWithBlock events matching the chain ID and optional filtering criteria.
 */
export function getValidRefundCandidates(chainId, hubPoolClient, spokePoolClients, filter, ignoredDepositValidationParams) {
    if (filter === void 0) { filter = {}; }
    if (ignoredDepositValidationParams === void 0) { ignoredDepositValidationParams = []; }
    return __awaiter(this, void 0, void 0, function () {
        var spokePoolClient, fromBlock, toBlock;
        var _this = this;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    spokePoolClient = spokePoolClients[chainId];
                    assert(isDefined(spokePoolClient));
                    fromBlock = filter.fromBlock, toBlock = filter.toBlock;
                    return [4 /*yield*/, mapAsync(spokePoolClient.getRefundRequests(), function (refundRequest) { return __awaiter(_this, void 0, void 0, function () {
                            var result, matchedDeposit;
                            return __generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0:
                                        if (isDefined(fromBlock) && fromBlock > refundRequest.blockNumber) {
                                            return [2 /*return*/, undefined];
                                        }
                                        if (isDefined(toBlock) && toBlock < refundRequest.blockNumber) {
                                            return [2 /*return*/, undefined];
                                        }
                                        return [4 /*yield*/, refundRequestIsValid(spokePoolClients, hubPoolClient, refundRequest, ignoredDepositValidationParams)];
                                    case 1:
                                        result = _a.sent();
                                        if (result.valid) {
                                            matchedDeposit = result.matchingDeposit;
                                            if (matchedDeposit === undefined) {
                                                throw new Error("refundRequestIsValid returned true but matchingDeposit is undefined");
                                            }
                                            return [2 /*return*/, __assign(__assign({}, refundRequest), { matchedDeposit: matchedDeposit })];
                                        }
                                        else {
                                            return [2 /*return*/, undefined];
                                        }
                                        return [2 /*return*/];
                                }
                            });
                        }); })];
                case 1: return [2 /*return*/, (_a.sent()).filter(function (refundRequest) { return refundRequest !== undefined; })];
            }
        });
    });
}
/**
 * Serializes a `UBAClientState` object.
 * @param ubaClientState `UBAClientState` object to serialize.
 * @returns Serialized `UBAClientState` object as a string.
 * @note The resulting string can be deserialized using `deserializeUBAClientState`.
 */
export function serializeUBAClientState(ubaClientState) {
    return stringifyJSONWithNumericString(ubaClientState);
}
/**
 * @todo Improve type safety and reduce `any`s.
 * Deserializes a serialized `UBAClientState` object.
 * @param serializedUBAClientState Serialized `UBAClientState` object that for example gets returned when calling `updateUBAClient` and serializing the result.
 * @returns Deserialized `UBAClientState` object. Specifically with `{ type: "BigNumber", value: "0x0" } converted to `BigNumber` instances. And a correct `UBAFeeConfig` instance.
 */
export function deserializeUBAClientState(serializedUBAClientState) {
    return Object.entries(serializedUBAClientState).reduce(function (acc, _a) {
        var _b;
        var chainId = _a[0], chainState = _a[1];
        if (typeof chainState !== "object") {
            throw new Error("Failed to parse chain state for chain ".concat(chainId));
        }
        return __assign(__assign({}, acc), (_b = {}, _b[chainId] = __assign(__assign({}, chainState), { bundles: Object.entries(chainState.bundles).reduce(function (acc, _a) {
                var _b;
                var tokenSymbol = _a[0], bundleStates = _a[1];
                if (!Array.isArray(bundleStates)) {
                    throw new Error("Failed to parse bundle states for token ".concat(tokenSymbol));
                }
                return __assign(__assign({}, acc), (_b = {}, _b[tokenSymbol] = bundleStates.map(function (bundleState) {
                    var deserializedUBAFeeConfig = deserializeUBAFeeConfig(bundleState.config);
                    return __assign(__assign({}, bundleState), { openingBalance: BigNumber.from(bundleState.openingBalance), openingIncentiveBalance: BigNumber.from(bundleState.openingIncentiveBalance), config: new UBAFeeConfig(deserializedUBAFeeConfig.baselineFee, deserializedUBAFeeConfig.balancingFee, deserializedUBAFeeConfig.balanceTriggerThreshold, deserializedUBAFeeConfig.lpGammaFunction, deserializedUBAFeeConfig.incentivePoolAdjustment, deserializedUBAFeeConfig.ubaRewardMultiplier), flows: bundleState.flows.map(deserializeModifiedUBAFlow) });
                }), _b));
            }, {}) }), _b));
    }, {});
}
function deserializeModifiedUBAFlow(serializedModifiedUBAFlow) {
    return {
        flow: deserializeBigNumberValues(serializedModifiedUBAFlow.flow),
        systemFee: deserializeBigNumberValues(serializedModifiedUBAFlow.systemFee),
        relayerFee: deserializeBigNumberValues(serializedModifiedUBAFlow.relayerFee),
        runningBalance: BigNumber.from(serializedModifiedUBAFlow.runningBalance),
        incentiveBalance: BigNumber.from(serializedModifiedUBAFlow.incentiveBalance),
        netRunningBalanceAdjustment: BigNumber.from(serializedModifiedUBAFlow.netRunningBalanceAdjustment),
    };
}
function deserializeUBAFeeConfig(serializedUBAFeeConfig) {
    var _a, _b, _c, _d, _e, _f;
    return {
        baselineFee: {
            default: BigNumber.from(serializedUBAFeeConfig.baselineFee.default),
            override: Object.entries((_a = serializedUBAFeeConfig.baselineFee.override) !== null && _a !== void 0 ? _a : {}).reduce(function (acc, _a) {
                var key = _a[0], value = _a[1];
                acc[key] = BigNumber.from(value);
                return acc;
            }, {}),
        },
        balancingFee: {
            default: serializedUBAFeeConfig.balancingFee.default.map(function (tuple) {
                return tuple.map(function (value) { return BigNumber.from(value); });
            }),
            override: Object.entries((_b = serializedUBAFeeConfig.balancingFee.override) !== null && _b !== void 0 ? _b : {}).reduce(function (acc, _a) {
                var key = _a[0], value = _a[1];
                acc[key] = value.map(function (tuple) { return tuple.map(function (value) { return BigNumber.from(value); }); });
                return acc;
            }, {}),
        },
        balanceTriggerThreshold: {
            default: {
                lowerBound: deserializeBigNumberValues(serializedUBAFeeConfig.balanceTriggerThreshold.default.lowerBound),
                upperBound: deserializeBigNumberValues(serializedUBAFeeConfig.balanceTriggerThreshold.default.upperBound),
            },
            override: Object.entries((_c = serializedUBAFeeConfig.balanceTriggerThreshold.override) !== null && _c !== void 0 ? _c : {}).reduce(function (acc, _a) {
                var key = _a[0], value = _a[1];
                acc[key] = {
                    lowerBound: deserializeBigNumberValues(value.lowerBound),
                    upperBound: deserializeBigNumberValues(value.upperBound),
                };
                return acc;
            }, {}),
        },
        lpGammaFunction: {
            default: serializedUBAFeeConfig.lpGammaFunction.default.map(function (tuple) {
                return tuple.map(function (value) { return BigNumber.from(value); });
            }),
            override: Object.entries((_d = serializedUBAFeeConfig.lpGammaFunction.override) !== null && _d !== void 0 ? _d : {}).reduce(function (acc, _a) {
                var key = _a[0], value = _a[1];
                acc[key] = value.map(function (tuple) { return tuple.map(function (value) { return BigNumber.from(value); }); });
                return acc;
            }, {}),
        },
        incentivePoolAdjustment: Object.entries((_e = serializedUBAFeeConfig.incentivePoolAdjustment) !== null && _e !== void 0 ? _e : {}).reduce(function (acc, _a) {
            var key = _a[0], value = _a[1];
            acc[key] = BigNumber.from(value);
            return acc;
        }, {}),
        ubaRewardMultiplier: Object.entries((_f = serializedUBAFeeConfig.ubaRewardMultiplier) !== null && _f !== void 0 ? _f : {}).reduce(function (acc, _a) {
            var key = _a[0], value = _a[1];
            acc[key] = BigNumber.from(value);
            return acc;
        }, {}),
    };
}
/**
 * Serializes a `UBABundleState` object.
 * @param obj The object to deserialize.
 * @returns An object with all BigNumber values converted to strings.
 */
function deserializeBigNumberValues(obj) {
    if (obj === void 0) { obj = {}; }
    return Object.entries(obj).reduce(function (acc, _a) {
        var _b, _c;
        var key = _a[0], value = _a[1];
        try {
            var parsedBigNumber = BigNumber.from(value);
            return __assign(__assign({}, acc), (_b = {}, _b[key] = parsedBigNumber, _b));
        }
        catch (_d) {
            // If throws then value is not a BigNumber.
            return __assign(__assign({}, acc), (_c = {}, _c[key] = value, _c));
        }
    }, {});
}
//# sourceMappingURL=UBAClientUtilities.js.map