import { __awaiter, __generator } from "tslib";
import assert from "assert";
import { fixedPointAdjustment, toBNWei, nativeToToken, toBN, min, max, percent, MAX_BIG_INT, } from "../utils";
export var expectedCapitalCostsKeys = ["lowerBound", "upperBound", "cutoff", "decimals"];
export var DEFAULT_LOGGER = {
    debug: function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        return console.debug(args);
    },
    info: function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        return console.info(args);
    },
    warn: function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        return console.warn(args);
    },
    error: function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        return console.error(args);
    },
};
var RelayFeeCalculator = /** @class */ (function () {
    function RelayFeeCalculator(config, logger, destinationChainId) {
        assert(config, "config must be provided");
        if ("queries" in config) {
            this.queries = config.queries;
        }
        else {
            assert(destinationChainId !== undefined, "destinationChainId must be provided if queriesMap is provided");
            assert(config.queriesMap[destinationChainId], "No queries provided for destination chain");
            this.queries = config.queriesMap[destinationChainId];
        }
        this.gasDiscountPercent = config.gasDiscountPercent || 0;
        this.capitalDiscountPercent = config.capitalDiscountPercent || 0;
        this.feeLimitPercent = config.feeLimitPercent || 0;
        this.nativeTokenDecimals = config.nativeTokenDecimals || 18;
        this.capitalCostsPercent = config.capitalCostsPercent || 0;
        assert(this.gasDiscountPercent >= 0 && this.gasDiscountPercent <= 100, "gasDiscountPercent must be between 0 and 100 percent");
        assert(this.capitalDiscountPercent >= 0 && this.capitalDiscountPercent <= 100, "capitalDiscountPercent must be between 0 and 100 percent");
        assert(this.feeLimitPercent >= 0 && this.feeLimitPercent <= 100, "feeLimitPercent must be between 0 and 100 percent");
        assert(this.capitalCostsPercent >= 0 && this.capitalCostsPercent <= 100, "capitalCostsPercent must be between 0 and 100 percent");
        this.capitalCostsConfig = {};
        for (var _i = 0, _a = Object.keys(config.capitalCostsConfig || {}); _i < _a.length; _i++) {
            var token = _a[_i];
            this.capitalCostsConfig[token] = RelayFeeCalculator.validateAndTransformCapitalCostsConfigOverride((config.capitalCostsConfig || {})[token]);
        }
        this.logger = logger || DEFAULT_LOGGER;
    }
    /**
     * Type guard to check if a config is a CapitalCostConfigOverride or a CapitalCostConfig.
     * @param config CapitalCostConfig or CapitalCostConfigOverride
     * @returns true if the config is a CapitalCostConfigOverride, false otherwise.
     * @private
     * @dev This is a type guard that is used to check if a config is a CapitalCostConfigOverride or a CapitalCostConfig.
     * This is needed because the config can be either a CapitalCostConfig or a CapitalCostConfigOverride. If it's a
     * CapitalCostConfig, then we need to convert it to a CapitalCostConfigOverride with the default config set with no route
     * overrides.
     */
    RelayFeeCalculator.capitalCostConfigIsOverride = function (config) {
        return config.default !== undefined;
    };
    /**
     * Validates a CapitalCostConfigOverride or a CapitalCostConfig.
     * @param capitalCosts CapitalCostConfig or CapitalCostConfigOverride
     * @returns CapitalCostConfigOverride
     */
    RelayFeeCalculator.validateAndTransformCapitalCostsConfigOverride = function (capitalCosts) {
        // We need to first convert the config to a baseline type. This is because the config can be either a CapitalCostConfig
        // or a CapitalCostConfigOverride. If it's a CapitalCostConfig, then we need to convert it to a CapitalCostConfigOverride with
        // the default config set with no route overrides.
        var config = this.capitalCostConfigIsOverride(capitalCosts)
            ? capitalCosts
            : { default: capitalCosts };
        // Validate the default config.
        this.validateCapitalCostsConfig(config.default);
        // Iterate over all the route overrides and validate them.
        for (var _i = 0, _a = Object.values(config.routeOverrides || {}); _i < _a.length; _i++) {
            var toChainIdRoutes = _a[_i];
            for (var _b = 0, _c = Object.values(toChainIdRoutes); _b < _c.length; _b++) {
                var override = _c[_b];
                this.validateCapitalCostsConfig(override);
            }
        }
        return config;
    };
    /**
     * Validates a CapitalCostConfig.
     * @param capitalCosts CapitalCostConfig
     */
    RelayFeeCalculator.validateCapitalCostsConfig = function (capitalCosts) {
        assert(toBN(capitalCosts.upperBound).lt(toBNWei("0.01")), "upper bound must be < 1%");
        assert(toBN(capitalCosts.lowerBound).lte(capitalCosts.upperBound), "lower bound must be <= upper bound");
        assert(capitalCosts.decimals > 0 && capitalCosts.decimals <= 18, "invalid decimals");
    };
    RelayFeeCalculator.prototype.getTokenPrice = function (tokenSymbol) {
        return this.queries.getTokenPrice(tokenSymbol);
    };
    RelayFeeCalculator.prototype.gasFeePercent = function (amountToRelay, tokenSymbol, _tokenPrice) {
        return __awaiter(this, void 0, void 0, function () {
            var getGasCosts, getTokenPrice, _a, gasCosts, tokenPrice, decimals, gasFeesInToken;
            var _this = this;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        if (toBN(amountToRelay).eq(0))
                            return [2 /*return*/, MAX_BIG_INT];
                        getGasCosts = this.queries.getGasCosts().catch(function (error) {
                            _this.logger.error({ at: "sdk-v2/gasFeePercent", message: "Error while fetching gas costs", error: error });
                            throw error;
                        });
                        getTokenPrice = this.queries.getTokenPrice(tokenSymbol).catch(function (error) {
                            _this.logger.error({ at: "sdk-v2/gasFeePercent", message: "Error while fetching token price", error: error });
                            throw error;
                        });
                        return [4 /*yield*/, Promise.all([
                                getGasCosts,
                                _tokenPrice !== undefined ? _tokenPrice : getTokenPrice,
                            ])];
                    case 1:
                        _a = _b.sent(), gasCosts = _a[0], tokenPrice = _a[1];
                        decimals = this.queries.getTokenDecimals(tokenSymbol);
                        gasFeesInToken = nativeToToken(gasCosts, tokenPrice, decimals, this.nativeTokenDecimals);
                        return [2 /*return*/, percent(gasFeesInToken, amountToRelay.toString())];
                }
            });
        });
    };
    // Note: these variables are unused now, but may be needed in future versions of this function that are more complex.
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    RelayFeeCalculator.prototype.capitalFeePercent = function (_amountToRelay, _tokenSymbol, _originRoute, _destinationRoute) {
        // If amount is 0, then the capital fee % should be the max 100%
        if (toBN(_amountToRelay).eq(toBN(0)))
            return MAX_BIG_INT;
        // V0: Charge fixed capital fee
        var defaultFee = toBNWei(this.capitalCostsPercent / 100);
        // V1: Charge fee that scales with size. This will charge a fee % based on a linear fee curve with a "kink" at a
        // cutoff in the same units as _amountToRelay. Before the kink, the fee % will increase linearly from a lower
        // bound to an upper bound. After the kink, the fee % increase will be fixed, and slowly approach the upper bound
        // for very large amount inputs.
        if (this.capitalCostsConfig[_tokenSymbol]) {
            var overrideConfig = this.capitalCostsConfig[_tokenSymbol];
            var config = overrideConfig.default;
            if (_originRoute && _destinationRoute && overrideConfig.routeOverrides) {
                var potentialDestinationRoutes = overrideConfig.routeOverrides[_originRoute];
                if (potentialDestinationRoutes) {
                    config = potentialDestinationRoutes[_destinationRoute] || config;
                }
            }
            // Scale amount "y" to 18 decimals.
            var y = toBN(_amountToRelay).mul(toBNWei("1", 18 - config.decimals));
            // At a minimum, the fee will be equal to lower bound fee * y
            var minCharge = toBN(config.lowerBound).mul(y).div(fixedPointAdjustment);
            // Charge an increasing marginal fee % up to min(cutoff, y). If y is very close to the cutoff, the fee %
            // will be equal to half the sum of (upper bound + lower bound).
            var yTriangle = min(config.cutoff, y);
            // triangleSlope is slope of fee curve from lower bound to upper bound. If cutoff is 0, slope is 0.
            // triangleCharge is interval of curve from 0 to y for curve = triangleSlope * y
            var triangleSlope = toBN(config.cutoff).eq(toBN(0))
                ? toBN(0)
                : toBN(config.upperBound).sub(config.lowerBound).mul(fixedPointAdjustment).div(config.cutoff);
            var triangleHeight = triangleSlope.mul(yTriangle).div(fixedPointAdjustment);
            var triangleCharge = triangleHeight.mul(yTriangle).div(toBNWei(2));
            // For any amounts above the cutoff, the marginal fee % will not increase but will be fixed at the upper bound
            // value.
            var yRemainder = max(toBN(0), y.sub(config.cutoff));
            var remainderCharge = yRemainder.mul(toBN(config.upperBound).sub(config.lowerBound)).div(fixedPointAdjustment);
            return minCharge.add(triangleCharge).add(remainderCharge).mul(fixedPointAdjustment).div(y);
        }
        return defaultFee;
    };
    RelayFeeCalculator.prototype.relayerFeeDetails = function (amountToRelay, tokenSymbol, tokenPrice, _originRoute, _destinationRoute) {
        return __awaiter(this, void 0, void 0, function () {
            var gasFeePercent, gasFeeTotal, capitalFeePercent, capitalFeeTotal, relayFeePercent, relayFeeTotal, maxGasFeePercent, minDeposit, isAmountTooLow;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.gasFeePercent(amountToRelay, tokenSymbol, tokenPrice)];
                    case 1:
                        gasFeePercent = _a.sent();
                        gasFeeTotal = gasFeePercent.mul(amountToRelay).div(fixedPointAdjustment);
                        return [4 /*yield*/, this.capitalFeePercent(amountToRelay, tokenSymbol, _originRoute, _destinationRoute)];
                    case 2:
                        capitalFeePercent = _a.sent();
                        capitalFeeTotal = capitalFeePercent.mul(amountToRelay).div(fixedPointAdjustment);
                        relayFeePercent = gasFeePercent.add(capitalFeePercent);
                        relayFeeTotal = gasFeeTotal.add(capitalFeeTotal);
                        maxGasFeePercent = max(toBNWei(this.feeLimitPercent / 100).sub(capitalFeePercent), toBN(0));
                        if (maxGasFeePercent.eq(toBN(0))) {
                            minDeposit = MAX_BIG_INT;
                            isAmountTooLow = true;
                        }
                        else {
                            minDeposit = gasFeeTotal.mul(fixedPointAdjustment).div(maxGasFeePercent);
                            isAmountTooLow = toBN(amountToRelay).lt(minDeposit);
                        }
                        return [2 /*return*/, {
                                amountToRelay: amountToRelay.toString(),
                                tokenSymbol: tokenSymbol,
                                gasFeePercent: gasFeePercent.toString(),
                                gasFeeTotal: gasFeeTotal.toString(),
                                gasDiscountPercent: this.gasDiscountPercent,
                                capitalFeePercent: capitalFeePercent.toString(),
                                capitalFeeTotal: capitalFeeTotal.toString(),
                                capitalDiscountPercent: this.capitalDiscountPercent,
                                relayFeePercent: relayFeePercent.toString(),
                                relayFeeTotal: relayFeeTotal.toString(),
                                feeLimitPercent: this.feeLimitPercent,
                                maxGasFeePercent: maxGasFeePercent.toString(),
                                minDeposit: minDeposit.toString(),
                                isAmountTooLow: isAmountTooLow,
                            }];
                }
            });
        });
    };
    return RelayFeeCalculator;
}());
export { RelayFeeCalculator };
//# sourceMappingURL=relayFeeCalculator.js.map