import { __awaiter, __generator } from "tslib";
import assert from "assert";
import { validateFillForDeposit } from "./FlowUtils";
import { getCurrentTime } from "./TimeUtils";
import { isDefined } from "./TypeGuards";
import { getDepositInCache, getDepositKey, setDepositInCache } from "./CachingUtils";
import { DEFAULT_CACHING_TTL } from "../constants";
// Load a deposit for a fill if the fill's deposit ID is outside this client's search range.
// This can be used by the Dataworker to determine whether to give a relayer a refund for a fill
// of a deposit older or younger than its fixed lookback.
/**
 * Attempts to resolve a deposit for a fill. If the fill's deposit Id is within the spoke pool client's search range,
 * the deposit is returned immediately. Otherwise, the deposit is queried first from the provided cache, and if it is
 * not found in the cache, it is queried from the spoke pool client. If the deposit is found, it is cached before
 * being returned.
 * @param spokePoolClient The spoke pool client to use to query the deposit.
 * @param fill The fill to resolve a deposit for.
 * @param cache An optional cache to use to store the deposit. Optional.
 * @returns The deposit for the fill, or undefined if the deposit could not be found.
 * @throws If the fill's origin chain ID does not match the spoke pool client's chain ID.
 * @throws If the spoke pool client has not been updated.
 */
export function queryHistoricalDepositForFill(spokePoolClient, fill, cache) {
    return __awaiter(this, void 0, void 0, function () {
        var deposit, cachedDeposit;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    if (fill.originChainId !== spokePoolClient.chainId) {
                        throw new Error("OriginChainId mismatch (".concat(fill.originChainId, " != ").concat(spokePoolClient.chainId, ")"));
                    }
                    // We need to update client so we know the first and last deposit ID's queried for this spoke pool client, as well
                    // as the global first and last deposit ID's for this spoke pool.
                    if (!spokePoolClient.isUpdated) {
                        throw new Error("SpokePoolClient must be updated before querying historical deposits");
                    }
                    if (fill.depositId < spokePoolClient.firstDepositIdForSpokePool ||
                        fill.depositId > spokePoolClient.lastDepositIdForSpokePool) {
                        return [2 /*return*/, undefined];
                    }
                    if (fill.depositId >= spokePoolClient.earliestDepositIdQueried &&
                        fill.depositId <= spokePoolClient.latestDepositIdQueried) {
                        return [2 /*return*/, spokePoolClient.getDepositForFill(fill)];
                    }
                    if (!cache) return [3 /*break*/, 2];
                    return [4 /*yield*/, getDepositInCache(getDepositKey(fill), cache)];
                case 1:
                    cachedDeposit = _a.sent();
                    _a.label = 2;
                case 2:
                    if (!isDefined(cachedDeposit)) return [3 /*break*/, 3];
                    deposit = cachedDeposit;
                    // Assert that cache hasn't been corrupted.
                    assert(deposit.depositId === fill.depositId && deposit.originChainId === fill.originChainId);
                    return [3 /*break*/, 6];
                case 3: return [4 /*yield*/, spokePoolClient.findDeposit(fill.depositId, fill.destinationChainId, fill.depositor)];
                case 4:
                    deposit = _a.sent();
                    if (!cache) return [3 /*break*/, 6];
                    return [4 /*yield*/, setDepositInCache(deposit, getCurrentTime(), cache, DEFAULT_CACHING_TTL)];
                case 5:
                    _a.sent();
                    _a.label = 6;
                case 6: return [2 /*return*/, validateFillForDeposit(fill, deposit) ? deposit : undefined];
            }
        });
    });
}
//# sourceMappingURL=DepositUtils.js.map